# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


def add_to_path():

    partial_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../../')
    workspace_path = os.path.abspath(partial_path)
    assert os.path.exists(workspace_path)

    projects = []

    for current, dirs, c in os.walk(str(workspace_path)):
        for dir in dirs:

            project_path = os.path.join(workspace_path, dir, 'biobroker')

            if os.path.exists(project_path):
                projects.append(project_path)

    for project_str in projects:
        sys.path.append(project_str)

add_to_path()


# -- Project information -----------------------------------------------------

project = 'Biobroker'
copyright = '2024, Enrique Sapena Ventura'
author = 'Enrique Sapena Ventura'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [ 'sphinx.ext.autodoc',
               'sphinx.ext.autosummary',
               'sphinx_autodoc_typehints',
               'sphinx_toolbox.decorators',
               'sphinx.ext.napoleon',
               'sphinxcontrib.mermaid',
               'myst_parser',
               'nbsphinx'
               ]

autosummary_generate = True
napoleon_google_docstring = False
napoleon_include_special_with_doc = True  # Document special members (like __str__) if they have documentation
napoleon_include_private_with_doc = True  # Document private members (Cls._private_member)
napoleon_use_param = False
napoleon_use_ivar = True

python_use_unqualified_type_names = True

add_module_names = False # Prettier documents?
autodoc_member_order = 'bysource' # I hate sorting - I decide where my methods appear >:(

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'biobroker.rst']


# -- Options for HTML output -------------------------------------------------
myst_fence_as_directive = ["mermaid"] # No longer need to use brackets for mermaid


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown'
}


from __future__ import annotations

from typing import TYPE_CHECKING

import os
import logging
import progressbar

# Need this to not hit a circular import issue with type hints
if TYPE_CHECKING:
    from biobroker.api import GenericApi
    from biobroker.authenticator import GenericAuthenticator
    from biobroker.input_processor import GenericInputProcessor
    from biobroker.output_processor import GenericOutputProcessor
    from biobroker.metadata_entity import GenericEntity
    from biobroker.wrangler.wrangler import Wrangler


progressbar.streams.wrap_stderr()  # This is needed for progressbar to work around loggers.


def set_up_logger(instance_object: GenericApi | GenericAuthenticator | GenericEntity | GenericInputProcessor |
                  GenericOutputProcessor | Wrangler, verbose: bool = False) -> logging.Logger:
    """
    Set up logger for any instance of a class.

    :param instance_object: Instance of the object being initialised.
    :param verbose: If true, set StreamHandler to INFO. If False, set to WARNING.
    :return: logging.Logger instance initialised to the class.
    """
    level = logging.INFO if verbose else logging.WARNING
    logger = logging.getLogger(instance_object.__class__.__name__)
    if logger.hasHandlers():
        return logger  # Avoid duplicating handlers when multiple instances of the same class are spawned
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if os.environ.get('BROKER_LOG'):
        file_handler = logging.FileHandler(os.environ.get('BROKER_LOG'))
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)  # File always set to level DEBUG
        logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(level)
    logger.addHandler(stream_handler)
    return logger



import os

import requests

from os.path import join
from requests.utils import requote_uri

from progressbar import progressbar, AdaptiveETA, Percentage, FormatLabel, AnimatedMarker, Counter, ProgressBar

from biobroker.api.exceptions import CantBeUpdatedApiError, CantBeUpdatedLocalError, ChecklistValidationError, \
    BiosamplesValidationError, BiosamplesNoErrorMessageError, StructuredDataError, StructuredDataSubmissionError
from biobroker.metadata_entity import Biosample
from biobroker.metadata_entity import GenericEntity
from biobroker.authenticator import GenericAuthenticator
from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.generic.logger import set_up_logger
from biobroker.generic.utilities import slice_list
from biobroker.generic.pydantic_model import StructuredDataModel

import pydantic_core


## TODO: BsdApi Build self url. Self points to general biosamples...


class GenericApi:
    """
    Generic API class. This class defines the minimal functions and class properties needed for the rest of the API
    classes.

    :param authenticator: Authenticator object. Requests are handled through the authenticator.
    :param base_uri: Base (root) uri of the API.
    :param verbose: Boolean indicating if the logger should be verbose.
    """
    def __init__(self, authenticator: GenericAuthenticator, base_uri: str, verbose: bool = True):
        self.authenticator = authenticator
        self.base_uri = base_uri
        self.logger = set_up_logger(self, verbose)

    def submit(self, entities: list[GenericEntity], **kwargs: dict) -> list[GenericEntity]:
        """
        Generic function for submitting an iterable of entities to the archive.

        :param entities: list of GenericEntity subclasses.
        :param kwargs: Keyword arguments needed for subclasses for submitting.
        :return: list of GenericEntity subclasses after archival/deposition.
        """
        if len(entities) > 1:
            return self._submit_multiple(entities, kwargs)
        else:
            return [self._submit(entities[0], kwargs)]

    def _submit(self, entity: GenericEntity, kwargs: dict) -> GenericEntity:
        """
        Generic function for submitting an entity to an archive.

        :param entity: Subclass of GenericEntity
        :param kwargs: Keyword arguments needed for subclasses' method.
        :return: Submitted GenericEntity subclass
        """
        raise MandatoryFunctionNotSet(self.logger)

    def _submit_multiple(self, entities: list[GenericEntity], kwargs: dict) -> list[GenericEntity]:
        """
        Generic function for submitting multiple entities to an archive.

        :param entities: List of subclasses of GenericEntity
        :param kwargs: Keyword arguments needed for subclasses' method.
        :return: Submitted GenericEntity subclasses
        """
        raise MandatoryFunctionNotSet(self.logger)

    def retrieve(self, accession: list[str]) -> list[GenericEntity]:
        """
        Generic function for retrieving one or more entities accessing the API via a/some unique identifier/s
        (accession). Depending on the type of input parameter, calls :func:`~GenericApi._retrieve` or
        :func:`~GenericApi._retrieve_multiple`.

        :param accession: Unique identifier for the entity. Can be a string or a list of strings.

        :return: List of entities retrieved from the API. MUST always return a list for consistency.
        """
        if not isinstance(accession, list):
            accession = [accession]
        if len(accession) > 1:
            return self._retrieve_multiple(accession)
        return [self._retrieve(accession[0])]

    def _retrieve(self, accession: str) -> GenericEntity:
        """
        Retrieve one entity via a unique identifier (accession).

        :param accession: Unique identifier for the entity.
        :return: An entity retrieved from the API.
        """
        raise MandatoryFunctionNotSet(logger=self.logger)

    def _retrieve_multiple(self, accession_list) -> list[GenericEntity]:
        """
        Retrieve multiple entities via a list of unique identifiers (accessions).

        :param accession_list: List of unique identifiers for the entities to retrieve.

        :return: List of entities.
        """
        raise MandatoryFunctionNotSet(logger=self.logger)

    def update(self, entity: list[GenericEntity]) -> list[GenericEntity]:
        """
        Update an entity. Should always take a list of entities as input, and each subclass decides how to handle
        the update.

        :param entity: List of GenericEntity's subclasses
        :return: List of updated GenericEntity's subclasses
        """
        if len(entity) > 1:
            return self._update_multiple(entity)
        return [self._update(entity[0])]

    def _update(self, entity: GenericEntity) -> GenericEntity:
        """
        Update an entity via the API.

        :param entity: GenericEntity subclass, corresponding to an entry in the database.
        :return: An updated GenericEntity subclass
        """
        raise MandatoryFunctionNotSet(logger=self.logger)

    def _update_multiple(self, entities: list[GenericEntity]) -> list[GenericEntity]:
        """
        Update multiple entities via the API.

        :param entities: list of GenericEntity subclasses, corresponding to several entries in the database.
        :return: list of updated GenericEntity's subclasses
        """
        raise MandatoryFunctionNotSet(logger=self.logger)


class BsdApi(GenericApi):
    """
    This is an API object specifically designed for the BioSamples Database.

    Please note: If you need to access the 'dev' environment, pleese set up the environment variable 'API_ENVIRONMENT'
    with the value 'dev'. Otherwise, this API object will point to the production BioSamples archive.

    :param authenticator: Subclass instance from the authenticator module. For BioSamples, it's recommended to use
                          the WebinAuthenticator.
    :param verbose: True if logger should be set to INFO. Default WARNING.
    """
    def __init__(self, authenticator: GenericAuthenticator, verbose: bool = True):
        environment = 'dev' if 'dev' == os.environ.get('API_ENVIRONMENT', '') else ''
        base_uri = "https://www.ebi.ac.uk/biosamples/samples".replace('www', f"www{environment}")
        super().__init__(authenticator, base_uri, verbose)
        self.bulk_accession_endpoint = join(self.base_uri.replace("biosamples/", "biosamples/v2/"), 'bulk-accession')
        self.bulk_submit_endpoint = join(self.base_uri.replace("biosamples/", "biosamples/v2/"), 'bulk-submit')
        self.validate_endpoint = join(self.base_uri, 'validate')
        self.structured_data_endpoint = self.base_uri.replace('/samples', '/structureddata')
        self.logger.info(f"Set up BSD API successfully: using base uri '{self.base_uri}'")
        self.relationship_types = ["derived_from", "same_as"]

    def _submit(self, entity: Biosample, kwargs: dict) -> Biosample:
        """
        Submit a single Biosample entity to BSD.

        :param entity: Biosample GenericEntity subclass.
        :param kwargs: Keyword argument. No use for this function.
        :return: a single, archived Biosample entity.
        """
        submit_url = self.base_uri
        r = self.authenticator.post(submit_url, payload=entity.entity)
        if r.status_code > 300:
            self._submit_errors(r)
        return Biosample(r.json())

    def _submit_multiple(self, entities: list[Biosample], kwargs: dict) -> list[Biosample]:
        """
        Submit a list of BioSample entities to biosamples, using the bulk-submit endpoint.

        :param entities: Iterable (List/Tuple) of BioSample objects. Must always be an iterable.
        :param kwargs: Keyword argument:

                       - 'chunk_size': integer, may be set up to determine the size of chunks to send to BSD at
                         once. Due to BSD technical limitations, capped at 500.
                       - 'process_relationships': bool, if set to true, after submission, updates the samples with
                         the relationships.

        :return: a list of BioSample entities
        """
        submission_results = []
        chunk_size = min(kwargs.get('chunk_size', 500), 500)
        self.logger.info(f"Submitting {len(entities)} samples to bulk endpoint: {self.bulk_submit_endpoint}")
        for entity_chunk in slice_list(entities, chunk_size):
            r = self.authenticator.post(self.bulk_submit_endpoint,
                                        payload=entity_chunk)
            if r.status_code > 300:
                self._submit_errors(r)
            results = r.json()
            submission_results.extend([Biosample(result) for result in results])

        if kwargs.get('process_relationships'):
            self.logger.info("Processing sample relationships. This may take a while.")
            submission_results = self.process_relationships(entities=submission_results)
        return submission_results

    # Retrieve/update/delete functions

    def _retrieve(self, accession: str) -> Biosample:
        """
        Retrieve a sample from BioSamples by using an accession

        :param accession: Accession ID, in BioSamples format
        :return: Biosample entity retrieved from the BioSample database
        """
        self.logger.info(f"Trying to retrieve sample with accession {accession}")
        return Biosample(self.authenticator.get(join(self.base_uri, accession)).json())

    def _retrieve_multiple(self, accession_list: list[str]) -> list[Biosample]:
        """
        Retrieve multiple samples from BioSamples by providing a list of accessions.

        :param accession_list: Iterable (tuple|list) with accessions
        :return: List of BioSample entities retrieved from BioSamples API
        """
        samples = [self._retrieve(accession) for accession in progressbar(accession_list,
                                                                          widgets=[FormatLabel('Retrieving samples: '),
                                                                                   Percentage(), " (", Counter(),
                                                                                   f"/{len(accession_list)}) ",
                                                                                   AnimatedMarker(markers='🀱🀲🀳🀴🀵🀶🀷🀾'
                                                                                                          '🁅🁌🁓🁚🁡'),
                                                                                   " ", AdaptiveETA()])]
        return samples

    def _update(self, entity: Biosample) -> Biosample:
        """
        Update a sample that is already in the BioSamples database. Samples must be updated using the FULL metadata, as
        per BSD specifications https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_update_sample

        :param entity: Biosample object loaded with the metadata, including the accession
        :return: Updated sample contained in Biosample object
        """

        is_invalid = BsdApi._is_invalid_for_update(entity)
        if is_invalid:
            raise CantBeUpdatedLocalError(sample_id=entity.id, reasons=is_invalid, logger=self.logger)
        sample_url = os.path.join(self.base_uri, entity.accession)
        response = self.authenticator.put(url=sample_url, payload=entity.entity)
        if response.status_code > 300:
            raise CantBeUpdatedApiError(sample_id=entity.id, response=response, logger=self.logger)
        return Biosample(response.json())

    def _update_multiple(self, entities: list[Biosample]) -> list[Biosample]:
        """
        Updates multiple samples in the BSD database. Since they can only be updated once at a time, calls
        :func:`~Biosample._update` once per sample in list.

        :param entities: List of Biosample entities to update
        :return: List with updated Biosample entities
        """
        return [self._update(entity) for entity in progressbar(entities,
                                                               widgets=[FormatLabel('Updating samples: '),
                                                                        Percentage(), " (", Counter(),
                                                                        f"/{len(entities)}) ",
                                                                        AnimatedMarker(markers='🀱🀲🀳🀴🀵🀶🀷🀾🁅🁌🁓🁚🁡'), " ",
                                                                        AdaptiveETA()])]

    # BioSamples-specific

    def validate_sample(self, entity: Biosample):
        """
        Validate a sample before submission. The errors returned are the same as the ones you get when you submit, so
        they are handled in the same way.

        :param entity: Biosample entity to be validated.
        :return:
        """
        r = self.authenticator.post(url=self.validate_endpoint, payload=entity.entity)
        self._submit_errors(r)

    def process_relationships(self, entities: list[Biosample]) -> list[Biosample]:
        """
        Process the relationships from a list of submitted entities. Assumes the relationships are defined in the
        metadata as `characteristics.derived_from/same_as`, and that the entities are linked via their `name`,
        not accession.

        If multiple relationships of the same type have to be defined, please use the :attr:`~biobroker.metadata_entity.Biosample.delimiter`
        as the input value (e.g. same_sample1||same_sample2 under `same_as` property)

        :param entities: List of Biosample entities to update their relationships.
        :return: list of updated entities
        """
        id_to_accession = {entity.id: entity.accession for entity in entities}
        for entity in entities:
            for relationship_type in self.relationship_types:
                if relationship_type in entity:
                    for split_relationship_value in entity[relationship_type]['text'].split(entity.delimiter):
                        target = split_relationship_value if Biosample.check_accession(split_relationship_value) \
                            else id_to_accession[split_relationship_value]
                        entity.add_relationship(source=entity.accession,
                                                target=target,
                                                relationship=relationship_type)
                    del entity[relationship_type]
        updated_entities = self.update(entities)
        return updated_entities

    def search_samples(self, text: str = "", attributes=None) -> list[Biosample]:
        """
        Search for samples in the Biosamples database. Can either search using free text (Can be improved using the
        query syntax specified here: https://www.ebi.ac.uk/ebisearch/documentation) or by attributes' values. For the
        attributes, please provide them as a dictionary.

        :param text: free text for the search. Can use query syntax for search engines (AND/OR etc)
        :param attributes: Attributes to filter by. Has to be provided as a dictionary {<attribute_name>: <attr. value>}
        :return: list of Biosamples or an empty list.
        """
        if attributes is None:
            attributes = dict()
        search_query = self._build_search_query(text, attributes)
        query_url = f"{self.base_uri}?{search_query}"

        response = self.authenticator.get(query_url).json()
        len_sample_search = response['page']['totalElements']
        size = response['page']['size']
        if not response.get('_embedded'):
            return []
        samples = response['_embedded']['samples']

        progress_bar = ProgressBar(widgets=[FormatLabel('Retrieving samples: '),
                                            Percentage(), " (", Counter(),
                                            f"/{len_sample_search}) ",
                                            AnimatedMarker(markers='🀱🀲🀳🀴🀵🀶🀷🀾🁅🁌🁓🁚🁡'), " ",
                                            AdaptiveETA()], max_value=len_sample_search)

        current = size
        while response['_links'].get('next'):
            response = self.authenticator.get(response['_links']['next']['href']).json()
            samples.extend(response['_embedded']['samples'])
            progress_bar.update(current)
            current += size
        progress_bar.finish()

        return [Biosample(sample) for sample in samples]

    def submit_structured_data(self, structured_data: dict) -> list[Biosample]:
        """
        Submit structured data to a sample in BioSamples. The data is checked before submission. May raise:
        - :exc:`~biobroker.api.exceptions.StructuredDataError`: Pre-submission errors
        - :exc:`~biobroker.api.exceptions.StructuredDataSubmissionError`: Post-submission errors

        :param structured_data: Structured data that's going to be posted in BSD. Must follow the format in https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_submit_structured_data
        :return: Biosample entity with the structured data
        """
        self._check_structured_data(structured_data=structured_data)
        structured_data_put_uri = join(self.structured_data_endpoint, structured_data['accession'])
        response = self.authenticator.put(url=structured_data_put_uri, payload=structured_data)
        if response.status_code == 200:
            return self.retrieve([structured_data['accession']])
        raise StructuredDataSubmissionError(self.logger, response)


    def _check_structured_data(self, structured_data: dict):
        """
        Check the structured data is correct using the data models and pydantic.
        Model used: :cls:`~biobroker.metadata_entity.data_model.StructuredDataModel`

        :param structured_data: Structured data.
        :raises: :exc:`~biobroker.api.exceptions.StructuredDataError`
        """
        try:
            StructuredDataModel.model_validate(structured_data)
        except pydantic_core.ValidationError as pydantic_error:
            raise StructuredDataError(logger=self.logger, errors=pydantic_error.errors())

    def _submit_errors(self, response: requests.Response) -> None:
        """
        Submission errors and how they should be handled. Biosamples returns non-jsonable responses sometimes so this
        handles the type and display of errors during submission.

        Errors being raised:
            - :exc:`~biobroker.api.exceptions.ChecklistValidationError` : Checklist validation has failed
            - :exc:`~biobroker.api.exceptions.BiosamplesValidationError` : BSD minimal sample checklist error. Returned differently, because why not

        :param response: response obtained during submission. Usually r.status_code > 300
        :return: None if no errors are detected.
        """
        if "Checklist validation failed" in response.text:
            raise ChecklistValidationError(response.text, self.logger)

        if response.status_code == 400:
            if "dataPath" in response.text:
                raise BiosamplesValidationError(response.text, self.logger)
            else:
                raise BiosamplesNoErrorMessageError(response.status_code, self.logger)

        return None

    @staticmethod
    def _build_search_query(text: str, attributes: dict) -> str:
        """
        Build the search query for BSD. Attributes need to be joined. Page=0 is specified to return pagination in the
        BioSamples API (Non-documented behaviour)

        :param text: Free text to search by.
        :param attributes: Dictionary of attributes and values to filter by.
        :return:
        """
        attributes_str = "&".join([f"filter=attr:{key}:{value}" for key, value in attributes.items()])
        query = f"text={text}&{attributes_str}&page=0"
        return requote_uri(query)

    @staticmethod
    def _is_invalid_for_update(entity: Biosample) -> list[str] | bool:
        """
        Checks if the sample is invalid for update.

        :param entity: Sample to be checked.
        :return: A list of validation errors or false
        """
        conditions = {
            "Accession not set in metadata": entity.accession,
            "Invalid accession format": entity.check_accession(entity.accession)
        }
        checks = [condition for condition, check in conditions.items() if not check]
        return checks if checks else False



from behave import *
import json
import inspect
import os
import sys
sys.path.insert(0, "../../")

from biobroker.api import GenericApi, BsdApi
from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.input_processor import TsvInputProcessor, XlsxInputProcessor, GenericInputProcessor
from biobroker.metadata_entity import GenericEntity, Biosample
from biobroker.output_processor import TsvOutputProcessor, XlsxOutputProcessor, GenericOutputProcessor
from biobroker.authenticator import WebinAuthenticator


def test_mandatory_methods(instance, list_of_methods):
    for mandatory_method in list_of_methods:
        try:
            method = eval(f"instance.{mandatory_method}")
            signature = inspect.signature(method)
            arguments = ["fakeValue" for _ in range(len(signature.parameters))]
            method(*arguments)
        except MandatoryFunctionNotSet:
            assert False, f"Please set mandatory method {mandatory_method} for subclass"
        except:
            continue

def GenericInputProcessor(input_processor: type[GenericInputProcessor]):
    assert input_processor.input_data

def GenericOutputProcessor(output_processor: type[GenericOutputProcessor]):
    assert output_processor.path

def GenericAuthenticator(authenticator: type[WebinAuthenticator]):
    assert authenticator.base_uri is not None
    assert authenticator.username is not None
    assert authenticator.password is not None
    assert authenticator.auth_endpoint is not None
    assert authenticator.token is not None

def GenericEntity(entity: type[GenericEntity]):
    assert entity.entity is not None
    assert entity.id is not None
    assert entity.accession is not None
    test_mandatory_methods(entity, ('flatten', '__setitem__', '__delitem__', '__contains__'))

def GenericApi(api: type[GenericApi]):
    assert api.authenticator is not None
    assert api.base_uri is not None
    test_mandatory_methods(api, ('_submit', '_sumit_multiple', '_retrieve', '_retrieve_multiple',
                               '_update', '_update_multiple'))


def load_credentials_webin():
    username = os.environ.get('WEBIN_USERNAME')
    password = os.environ.get('WEBIN_PASSWORD')
    assert username and password, f"Username or password missing from environment variables for authenticator with prefix WEBIN"
    return username, password

def load_webin_authenticator():
    return (WebinAuthenticator(*load_credentials_webin()),)

def load_biosample_valid_json():
    with open('assets/valid_minimal.json', 'r') as f:
        value = json.load(f)
    return (value,)


@given('I have a {subclass_object} and its init arguments: {arguments}')
def step_impl(context, subclass_object, arguments):
    context.entity = eval(subclass_object)
    context.args = eval(arguments)
    assert isinstance(context.args, list), "Arguments must be provided as a list"

@when('I load the instance with the arguments')
def step_impl(context):
    # Checking for functions
    load_arguments = []
    for index, arg in enumerate(context.args):
        match arg:
            case "function":
                function = eval(context.args[index + 1])
                load_arguments.extend(function())
                del context.args[index + 1]
            case _:
                load_arguments.append(arg)

    context.loaded_instance = context.entity(*load_arguments)

@then('it should have all mandatory fields and methods from the parent class {parent_class}')
def step_impl(context, parent_class):
    eval(parent_class)(context.loaded_instance)



import logging
from biobroker.generic.utilities import parse_pydantic_errors

class EntityValidationError(Exception):
    def __init__(self, logger: logging.Logger, entity_id: str, errors: list[dict]):
        delimiter = "\n\t- "
        self.message = (f"Metadata content has failed validation for '{entity_id}':"
                        f"{delimiter}{delimiter.join(parse_pydantic_errors(errors))}")
        logger.error(self.message)
        super().__init__(self.message)

class NoNameSetError(Exception):
    """Name has not been set in the sample"""
    def __init__(self, logger: logging.Logger, sample_id: str):
        self.message = f"Property 'name' needs to be set-up for the sample '{sample_id}'"
        logger.error(self.message)
        super().__init__(self.message)


class NameShouldBeStringError(Exception):
    """Name should be a string, no other types allowed"""
    def __init__(self, logger: logging.Logger, name):
        self.message = f"Property 'name' should be a string. Current value: {name}"
        logger.error(self.message)
        super().__init__(self.message)


class RelationshipInvalidSourceError(Exception):
    """Invalid source for relationship"""
    def __init__(self, logger: logging.Logger, sample_id: str, source: str):
        self.message = (f"Sample {sample_id}: invalid relationship source '{source}'. Only Biosample accessions can be "
                        "used")
        logger.error(self.message)
        super().__init__(self.message)


class RelationshipInvalidTargetError(Exception):
    """Invalid target for relationship"""
    def __init__(self, logger: logging.Logger, sample_id: str, target: str):
        self.message = (f"Sample {sample_id}: invalid relationship target '{target}'. Only Biosample accessions can be "
                        "used")
        logger.error(self.message)
        super().__init__(self.message)

class NoOrganismSetError(Exception):
    def __init__(self, logger: logging.Logger, sample_id: str):
        self.message = f"Property 'organism' (Or any variant) needs to be set-up for the sample '{sample_id}'"
        logger.error(self.message)
        super().__init__(self.message)



import logging
import json
from requests import Response

from json.decoder import JSONDecodeError

from biobroker.generic.utilities import parse_pydantic_errors

def parse_checklist_validation_errors(validation_errors: list[dict]):
    """
    Parse a checklist validation error and return it in a printable state.

    :param validation_errors: Response list with validation error dictionaries, containing `dataPath` and `errors`.
    :return: printable string.
    """
    validation_errors_str = "\n\t- ".join([f"{validation_errors[i]['dataPath'].replace('/characteristics/', '')}: "
                                           f"{','.join(validation_errors[i]['errors'])}"
                                           for i in range(len(validation_errors))])
    return validation_errors_str


class AccessionHasIncorrectFormat(Exception):
    def __init__(self, accession: str, logger: logging.Logger):
        """
        Raise when accession has incorrect format.

        :param accession: string passed as an accession
        :param logger: subclasss logger to log the error message to.
        """
        message = f"Could not retrieve accession {accession}; Incorrect format"
        logger.error(message)
        super().__init__(message)


class CantBeUpdatedApiError(Exception):
    def __init__(self, sample_id, response: Response, logger: logging.Logger):
        """
        Raise when an error comes from an API update request.

        :param sample_id: Id of the sample that raised the error
        :param response: errored response (r.status_code >= 300)
        :param logger: subclasss logger to log the error message to.
        """
        try:
            message = f"Sample with ID {sample_id} can't be updated:\n\t-Error type: {response.json()['error']}" \
                      f"\n\t-Error message: {response.json()['message']}"
        except TypeError:
            # Can't be updated, but response does not contain "error" dataPath
            message = f"Sample with ID {sample_id} can't be updated:\n\t- {parse_checklist_validation_errors(response.json())}"
        except JSONDecodeError:
            # Validation against checklist failed, returns a non-jsonable message
            validation_errors = json.loads(response.text.split('failed: ')[2])
            validation_errors_str = parse_checklist_validation_errors(validation_errors)
            message = f"Sample with ID {sample_id} failed update due to checklist errors:\n\t- {validation_errors_str}"
        logger.error(message)
        super().__init__(message)


class CantBeUpdatedLocalError(Exception):
    def __init__(self, sample_id: str, reasons: list[str], logger: logging.Logger):
        """
        Raise when an error comes from an update request, before doing the request.

        :param sample_id: ID of the sample that fail validation.
        :param reasons: list of reasons (strings) as to why validation failed.
        :param logger: subclass logger to log the error message to.
        """
        reasons_str = "\n\t- ".join(reasons)
        self.message = f"Sample with ID {sample_id} can't be updated - Error(s): \n\t- {reasons_str}"
        logger.error(self.message)
        super().__init__(self.message)


class BiosamplesValidationError(Exception):
    def __init__(self, response_text: str, logger: logging.Logger):
        """
        Raise a Biosamples minimal checklist validation error
        (https://www.ebi.ac.uk/biosamples/schemas/certification/biosamples-minimal.json)

        :param response_text: text from the response.
        :param logger: subclass logger to log the error message to.
        """
        validation_errors = json.loads(response_text)
        errors_str = parse_checklist_validation_errors(validation_errors)
        self.message = f"Found following errors in sample validation:\n\t- {errors_str})"
        logger.error(self.message)
        super().__init__(self.message)


class BiosamplesNoErrorMessageError(Exception):
    def __init__(self, status_code: int, logger: logging.Logger):
        """
        Raise an error without any other issue but a 400 in the response and no text whatsoever.
        Current testing (Completely empirical) has told me that this can be raised in the following situations:
            - `release` in incorrect format

        :param logger: subclass logger to log the error message to.
        """
        self.message = (f"It's dangerous to go alone, take this error code: '{status_code}' with you (There was no "
                        f"error message))")
        logger.error(self.message)
        super().__init__(self.message)


# I know this is repeated in the update error, but BSD does not return the sample ID with the creation of the sample.
# TODO: Future possible update: raise this error from cantbeupdated, with optional sample parameter. Right now not needed.
class ChecklistValidationError(Exception):
    def __init__(self, response_text: str, logger: logging.Logger):
        """
        Raise a checklist validation error.

        :param response_text: Text of the response.
        :param logger: subclass logger to log the error message to.
        """
        # Validation against checklist failed, returns a non-jsonable message
        validation_errors = json.loads(response_text.split('failed: ')[2])
        validation_errors_str = parse_checklist_validation_errors(validation_errors)
        self.message = f"Samples failed checklist validation with the following errors: \n\t- {validation_errors_str}"
        logger.error(self.message)
        super().__init__(self.message)


class StructuredDataError(Exception):
    def __init__(self, logger: logging.Logger, errors: list[dict]):
        delimiter = "\n\t- "
        self.message = (f"Structured data submission failed due to the following error(s):"
                        f"{delimiter}{delimiter.join(parse_pydantic_errors(errors))}\n Please see "
                        f"https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_submit_structured_data for more "
                        f"information")
        logger.error(self.message)
        super().__init__(self.message)

class StructuredDataSubmissionError(Exception):
    def __init__(self, logger: logging.Logger, response: Response):
        self.message = f"Error submitting structured data: {response.text}"
        logger.error(self.message)
        super().__init__(self.message)


from behave import *
import json

import sys
sys.path.insert(0, "../../")

from biobroker.metadata_entity import Biosample

@given("all the metadata entity classes")
def metadata_entity(context):
    context.instances = {}
    for subclass_entity, metadata_json_path in (('Biosample', 'assets/accessioned_BsdApi_entity.json'),):
        subclass_entity_object = eval(subclass_entity)
        with open(metadata_json_path, 'r') as f:
            context.instances[subclass_entity] = subclass_entity_object(json.load(f))


@given("a {metadata_entity}")
def flatten(context, metadata_entity):
    context.metadata_entity_name = metadata_entity
    context.metadata_entity = context.instances[metadata_entity]

@when("the flatten method is called")
def flatten(context):
    context.flattened_return = context.metadata_entity.flatten()

@then("it should result in a non-nested dictionary")
def flatten(context):
    assert all([not isinstance(value, dict) for value in context.flattened_return.values()])

@when("I add a set of key:value pairs")
def add_values(context):
    context.metadata_entity['key'] = 'value'
    context.metadata_entity['status'] = 'PUBLIC'

@then("they should be loaded in the entity in the proper place")
def add_values(context):
    match context.metadata_entity_name:
        case "Biosample":
            assert context.metadata_entity.entity['characteristics'].get('key') == [{'text': 'value'}]
            assert context.metadata_entity.entity.get('status') == 'PUBLIC'
        case _:
            assert False, f"This test needs to be set up for {context.metadata_entity_name}"

@when("I delete a value")
def delete_values(context):
    del context.metadata_entity['release']

@then("it should be deleted from the entity")
def delete_values(context):
    match context.metadata_entity_name:
        case "Biosample":
            assert context.metadata_entity.entity.get('release') is None
        case _:
            assert False, f"This test needs to be set up for {context.metadata_entity_name}"

@when("I add a valid relationship")
def valid_relationship(context):
    context.metadata_entity.add_relationship(source="SAMEA131439753", target="SAMEA1234533", relationship="derived_from")

@then("it should be loaded at the root, under 'relationships'")
def valid_relationship(context):
    assert isinstance(context.metadata_entity.entity.get('relationships'), list)
    assert context.metadata_entity['relationships'] == [{
            "source": "SAMEA131439753",
            "target": "SAMEA1234533",
            "type": "derived_from"
        }]

@when("I add a valid external reference")
def valid_external_reference(context):
    external_reference = "https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_example_1"
    context.metadata_entity.add_external_reference(external_reference)

@then("it should be loaded at the root, under 'external_references'")
def valid_external_reference(context):
    assert isinstance(context.metadata_entity.entity.get('externalReferences'), list)
    assert context.metadata_entity['externalReferences'] == [{"url": "https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_example_1"}]


from copy import deepcopy
from typing import Type

from pandas import read_csv, read_excel
from numpy import nan

from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.generic.logger import set_up_logger
from biobroker.metadata_entity import GenericEntity

"""
Test the behave tests
"""

class GenericInputProcessor:
    """
    Generic input processor.

    :param input_data_path: Path to the input file.
    :param verbose: Boolean indicating if the logger should be verbose.
    """
    def __init__(self, input_data_path: str, verbose: bool = False):
        self._input_data = ''
        self.logger = set_up_logger(self, verbose=verbose)
        self.input_data = input_data_path

    @property
    def input_data(self) -> list[dict]:
        """
        Input data in JSON format. Set from input_path by setter

        :return: Input data in JSON format
        """
        return self._input_data

    @input_data.setter
    def input_data(self, path: str) -> list[dict]:
        """
        SUBCLASSES MUST OVERRIDE THIS PROPERTY.

        Must assign "self._input_data" input for the "process" function to use.
        """
        raise MandatoryFunctionNotSet(self.logger)

    def process(self, entity: Type[GenericEntity]) -> list[GenericEntity]:
        """
        Process self.input_data and return a list of metadata entities that depend on the 'GenericEntity' subclass
        passed to the function.

        Any field with value "None" should be processed by the entity type; different services may require different
        behaviours.

        :param entity: GenericEntity subclass (Not instance) to process the input data into.
        :return: list of entities. Must be subclass of GenericEntity
        """
        entities = []
        for json_entity in self.input_data:
            new_entity = entity(metadata_content=deepcopy(json_entity))
            entities.append(new_entity)
        return entities

    def transform(self, field_mapping: dict, delete_non_mapped_fields: bool = False):
        """
        Transform the input data retrieved from the source to have new keys based on an {<old_key>: <new_key>} map. Only
        simple key name change operations are allowed.

        :param field_mapping: {'old_key': 'new_key'} dictionary. Not nested.
        :param delete_non_mapped_fields: Boolean to indicate if fields not present in the map above should be deleted.
               Defaults False
        """
        for i, entity in enumerate(self.input_data):
            for key, value in field_mapping.items():
                try:
                    field_value = entity.pop(key)
                except KeyError:
                    self.logger.warning(f"'transform' function: Key {key} was not found in entity with index {i}. "
                                        "This entity will not be updated")
                    break
                entity[value] = field_value
            if delete_non_mapped_fields:
                [entity.pop(key) for key in entity.keys() if key not in field_mapping.keys()]


class TsvInputProcessor(GenericInputProcessor):
    """
    TSV input processor. Loads a TSV file with entity metadata.

    :param input_data: Path to the file with the input metadata
    :param field_mapping: Path to the file with the field mapping
    """
    def __init__(self, input_data):
        super().__init__(input_data)

    @GenericInputProcessor.input_data.setter
    def input_data(self, path: str):
        """
        Setter for input_data property. Reads a TSV file with the pandas library, and returns a list with JSON files.

        :param path: Path to input data
        """
        file = read_csv(path, sep="\t", encoding='cp437').fillna(nan).replace([nan], [None])
        json_file = file.to_dict(orient='records')
        self._input_data = json_file


class XlsxInputProcessor(GenericInputProcessor):
    """
    XLSX input processor. Loads a XLSX file with entity metadata.

    :param input_data: Path to the file with the input metadata.
    :param worksheet_name: Name of the worksheet to be processed.
    """
    def __init__(self, input_data: str, sheet_name: str = "Sheet1"):
        self.sheet_name = sheet_name
        super().__init__(input_data)

    @GenericInputProcessor.input_data.setter
    def input_data(self, path):
        """
        Setter for the input_data property. Reads an xlsx file, on a specific worksheet name.

        :param path: Path to the file with the input metadata.
        """
        file = read_excel(path, engine='openpyxl', sheet_name=self.sheet_name).fillna(nan).replace([nan], [None])
        json_file = file.to_dict(orient='records')
        self._input_data = json_file


"""
API module. This module consists of several classes whose function is to provide with the necessary endpoints
for interacting with the archival services.

The main functions are to submit, retrieve with an accession, and update with an accession; However, as many
methods can be added as necessary.

**Mandatory arguments**:

- authenticator: Subclass of GenericAuthenticator
- base_uri: base uri for the archive's API (Can be automatically set-up by subclasses, see BsdApi object for an example)

**Optional arguments**:

- verbose: set to `True` if you want `INFO` and above-level logging events. If not set or set to False, only `WARNING`
  and above will be displayed

**Environment variables**:

- API_ENVIRONMENT: Needs to be set up if you want to set up a 'dev' authenticator. Please note this
  environment variable is shared with the Authenticator: this is to avoid inconsistent API/Authenticator combos (And
  even with all these checks and constraints, there will be errors, I'm pretty sure)

**Subclasses of GenericApi must define the following methods/properties**:

- _submit: Function called by `submit` when only one entity is sent to `submit`
- _submit_multiple: Function called by `submit` when multiple entities are sent to `submit`
- Same with `retrieve` and `update`.
"""

from .api import GenericApi, BsdApi

# This lets Sphinx know you want to document package.module.Class as package.Class.
__all__ = ['GenericApi', 'BsdApi']


"""
Input metadata processor. The goal of this module is to take an input file, read it and transform it into a list of
dictionaries containing the metadata for the different entities. Once in a flattened JSON state, they can be processed
into all the different entities.

**Mandatory arguments**:

- input_data: Path to the input file or loaded content. To be decided by subclass processor.

**Optional arguments**:

- verbose: set to `True` if you want `INFO` and above-level logging events. If not set or set to False, only `WARNING`
  and above will be displayed

**Subclasses of GenericInputProcessor must define the following methods/properties**:

- @input_data.setter

**Aspects to improve**:

- Currently, the :func:`~biobroker.input_processor.GenericInputProcessor.process`
  function can fail at any point if any entity fails to validate. This could be handled in 2 ways:

    - Catch all exceptions (meh) log them, and create the rest of the entities.
    - `Current behaviour`: fail miserably and not return anything. I really like this option as, for me, an input
      (spreadsheet, tsv file, etc) probably has meaning together and should remain this way. It could
      be improved by logging all errors, then failing.
"""

from .input_processor import GenericInputProcessor, TsvInputProcessor, XlsxInputProcessor
__all__ = ['GenericInputProcessor', 'TsvInputProcessor', 'XlsxInputProcessor']



"""
Output metadata processor. The goal of this module is to take an input list of GenericEntity's, transform it into a
dataframe and save it with pandas functionality into different formats. Pretty simple!

**Mandatory arguments**:

- output_path: Path to the file to save the metadata.

**Optional arguments**:

- verbose: set to `True` if you want `INFO` and above-level logging events. If not set or set to False, only `WARNING`
           and above will be displayed

**Subclasses of GenericOutputProcessor must define the following methods/properties**:

- _save
"""

from .output_processor import GenericOutputProcessor, TsvOutputProcessor, XlsxOutputProcessor

__all__ = ['GenericOutputProcessor', 'TsvOutputProcessor', 'XlsxOutputProcessor']



from typing import Generator


def slice_list(list_to_chunk: list | tuple, chunk_size: int) -> Generator:
    """
    Slice an iterable (List/Tuple) returning tuples of 'chunk_sixe' size.

    :param list_to_chunk: iterable to slice
    :param chunk_size: Size of the chunk
    :return: Generator of 'chunk_size' sizes of the list.
    """
    n = max(1, chunk_size)
    return (list_to_chunk[i:i+n] for i in range(0, len(list_to_chunk), n))


def parse_pydantic_errors(pydantic_errors: list[dict]) -> list:
    messages = []
    for error in pydantic_errors:
        location = "-->".join([str(e) for e in error['loc']]) if error['loc'] else "root"
        user_input = error['input']
        match error['type']:
            case 'model_type':
                message = "Value provided should be a dictionary"
            case 'string_pattern_mismatch':
                message = f"Value provided must match pattern {error['ctx']['pattern']}"
            case 'missing':
                message = f"Missing mandatory field '{error['loc'][-1]}'"
                location = "-->".join([str(e) for e in error['loc']][:-1]) if len(error['loc']) > 1 else "root"
            case _:
                message = error['msg']
        messages.append(f"{location}: {message}. Provided value: '{user_input}'")
    return messages


import pandas

from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.generic.logger import set_up_logger
from biobroker.metadata_entity import GenericEntity


class GenericOutputProcessor:
    """
    Generic output processor. Defines the mandatory functions for the subclasses to function.

    :param output_path: path to save the file. Please include the name and extension of the file.
    """
    def __init__(self, output_path: str, verbose: bool = False):
        self.logger = set_up_logger(self, verbose=verbose)
        self.path = output_path

    def save(self, entities: list[GenericEntity]):
        """
        Transform the entities into a dataframe to use pandas functionality to save.

        :param entities: Subclasses of GenericEntity.
        """
        json_to_save = [entity.flatten() for entity in entities]
        dataframe = pandas.DataFrame(json_to_save)
        self._save(dataframe)

    def _save(self, dataframe: pandas.DataFrame):
        """
        Function to be overriden by subclasses. Takes a dataframe and saves the output into self.path.

        :param dataframe: Dataframe containing the flattened metadata from the GenericEntity subclasses.
        """
        raise MandatoryFunctionNotSet(logger=self.logger)


class TsvOutputProcessor(GenericOutputProcessor):
    """
    TSV output processor. Takes a list of entities and outputs a TSV with the metadata processed.

    :param output_path: Path to the file being saved. Please include tsv extension.
    """
    def __init__(self, output_path: str):
        super().__init__(output_path)

    def _save(self, dataframe: pandas.DataFrame):
        """
        Save the resulting dataframe from :func:`~GenericOutputProcessor.save` into a tsv,
        using pandas functionality. NO, the delimiter is not customizable. Create another subclass if you want that.
        TSV means `TAB-Separated Values`, not comma, not pipes, not anything else. You weirdo.

        :param dataframe: Dataframe containing the flattened metadata from the GenericEntity subclasses.
        """
        separator = '\t'
        dataframe.to_csv(self.path, sep=separator, index=False)


class XlsxOutputProcessor(GenericOutputProcessor):
    """
    Excel output processor. Takes a list of entities and outputs an excel file with the metadata processed.

    :param output_path: Path to the file being saved. Please include '.xlsx' extension.
    """
    def __init__(self, output_path, sheet_name: str = 'Sheet1'):
        super().__init__(output_path)
        self.sheet_name = sheet_name

    def _save(self, dataframe: pandas.DataFrame):
        """
        Save the resulting dataframe from :func:`~GenericOutputProcessor.save` into an excel.

        :param dataframe: Dataframe containing the flattened metadata from the GenericEntity subclasses.
        """
        dataframe.to_excel(self.path, index=False, sheet_name=self.sheet_name, engine='openpyxl')



import os

def remove_generated_files(paths: list):
    for file in paths:
        os.remove(file)


def before_feature(context, feature):
    if "integration" in feature.tags:
        feature.skip("Integration tests have to be triggered explicitly")
        return


def after_scenario(context, scenario):
    match scenario.name:
        case x if x.startswith("Save metadata to source --"):
            remove_generated_files([context.output_file_path])




"""
Authenticator module. This module consists of several classes that can load a username and a password
and retrieve the token needed to validate the requests.

For easiness of use, it also wraps the requests, adding automatically the token to the header, and adding an
extra check in case the token expires.

**Mandatory arguments**:

- username: Username for the authentication service
- password: Password for the authentication service
- base_uri: base uri for the authentication service (Can be provided by subclasses, see WebinApi object for an example)

**Optional arguments**:

- verbose: set to `True` if you want `INFO` and above-level logging events. If not set or set to False, only `WARNING`
  and above will be displayed

**Environment variables**:

- API_ENVIRONMENT: Needs to be set up if you want to set up a 'dev' authenticator. Please note this
  environment variable is shared with the API: this is to avoid inconsistent API/Authenticator combos (And even with
  all these checks and constraints, there will be errors, I'm pretty sure)

**Subclasses of GenericAuthenticator must define the following methods/properties**:

- @token.setter

**Aspects to improve**:

- _request function: Currently, it assumes JWT token generation + authorization "Bearer <token>" in all subclasses.
  if other headers or payloads are necessary for authentication, it needs to be overridden.
"""

from .authenticator import GenericAuthenticator, WebinAuthenticator

__all__ = ['GenericAuthenticator', 'WebinAuthenticator']


import datetime
from enum import Enum
from dateutil import parser
from typing import Dict, Optional

from pydantic import BaseModel, Field, field_validator, ConfigDict

"""
FIELDS/SUPPORTING MODELS
------------------------
"""


class CharacteristicsFields(BaseModel):
    model_config = ConfigDict(extra='forbid')
    text: str | datetime.datetime | int | float
    unit: Optional[str] = None
    ontologyTerms: Optional[list[str]] = None
    tag : Optional[str] = None

    @field_validator('text')
    @classmethod
    def proper_formats(cls, value):
        """
        Evaluate if an input is a date/int/float and give it a proper formatting, returning it as str
        """
        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%dT%H:%M:%SZ")
            if value.endswith('T00:00:00Z'):
                value = value.replace('T00:00:00Z', '')
            return value
        elif isinstance(value, float):
            return str(int(value)) if value.is_integer() else str(value)
        return str(value)

class DataContent(BaseModel):
    value: str

class DataEntry(BaseModel):
    webinSubmissionAccountId: str = Field(pattern="Webin-[0-9]+$")
    type: str
    content: list[Dict[str, DataContent]]

class RelationshipType(str, Enum):
    derived_from = "derived_from"
    same_as = "same_as"
    has_member = "has_member"
    child_of = "child_of"

class Relationship(BaseModel):
    model_config = ConfigDict(extra='forbid')
    target: str = Field(pattern="^SAMEA[0-9]+$")
    source: str = Field(pattern="^SAMEA[0-9]+$")
    type: RelationshipType

class Organization(BaseModel):
    Name: str

class ExternalUrl(BaseModel):
    url: str = Field(pattern="https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}([-a-zA-Z0-9()@:%_\+.~#?&\/=]*)")
    doi: Optional[str] = None

"""
DATA MODELS
-----------
"""

class StructuredDataModel(BaseModel):
    """
    Biosamples' structured data model
    """
    accession: str = Field(pattern="^SAME.[0-9]+$")
    data: list[DataEntry]

class BiosampleGeneralModel(BaseModel):
    """
    Biosamples General model
    """
    model_config = ConfigDict(extra='allow')
    name: str = Field(min_length=1)
    accession: Optional[str] = None
    release: str
    characteristics: Dict[str, list[CharacteristicsFields]]
    relationships: Optional[list[Relationship]] = None
    organization: Optional[list[Organization]] = None
    structuredData: Optional[list[DataEntry]] = None
    externalReferences: Optional[list[ExternalUrl]] = None

    @field_validator('release')
    @classmethod
    def parse_date(cls, value):
        try:
            value = parser.isoparse(value).strftime("%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            raise ValueError("Invalid date format. Should be provided as YYYY-MM-DD with optional Thh:mm:ss.sssZ") from None
        return value

    @field_validator('characteristics')
    @classmethod
    def organism_must_be_set(cls, value):
        valid_organism_keys = ('organism', 'Organism', 'species', 'Species')
        try:
            which_one = [key in value for key in valid_organism_keys]
            which_one.index(True)
        except ValueError:
            raise ValueError("'organism' must be set. Please use the keys 'organism', 'Organism', 'species' or 'Species'") from None
        value['organism'] = value[valid_organism_keys[which_one.index(True)]]
        for organism_key in valid_organism_keys[1:]:
            if organism_key in value:
                del value[organism_key]
        return value


from typing import Type

from biobroker.api import GenericApi
from biobroker.authenticator import GenericAuthenticator
from biobroker.input_processor import GenericInputProcessor
from biobroker.metadata_entity import GenericEntity
from biobroker.output_processor import GenericOutputProcessor

from biobroker.generic.logger import set_up_logger


class Wrangler:
    """
    Wrapper submodule! Provides with the very basic functionality expected for all subclasses of all the generic
    submodules defined in the library. Currently allows:

    - Submitting a set of entities to an archive
    - Retrieving a set of entities from an archive
    - Saving a set of entities to a file

    And the functionality depends on what submodule you create it with!

    :param api: Subclass of :mod:`broker.api.generic_api.GenericApi`
    :param authenticator: Subclass of :mod:`broker.authenticator.generic_authenticator.GenericAuthenticator`
    :param input_processor: Subclass of :mod:`broker.input_processor.generic_input_processor.GenericInputProcessor`
    :param entity: Subclass of :mod:`broker.metadata_entity.generic_entity.GenericEntity` (Please notice, you need to
                   pass a class to this parameter)
    :param output_processor: subclass of :mod:`broker.output_processor.generic_output_processor.GenericOutputProcessor`
    """
    def __init__(self, api: GenericApi, authenticator: GenericAuthenticator, input_processor: GenericInputProcessor,
                 entity: Type[GenericEntity], output_processor: GenericOutputProcessor):
        self.logger = set_up_logger(instance_object=self, verbose=True)
        self.authenticator = authenticator
        self.input_processor = input_processor
        self.api = api
        self.output_processor = output_processor
        self.entities = self.input_processor.process(entity=entity)

    def submit(self, entities: list[GenericEntity]) -> list[GenericEntity]:
        """
        Submit the entities to the archive.

        :param entities: List of genericEntity's.

        :return: List of entities submitted
        """
        return self.api.submit(entities)

    def save_results(self, entities):
        """
        Save the entities to a file.

        :param entities: Entities to be saved to a file
        """
        self.output_processor.save(entities)

    def retrieve_entities(self, accession_list: list[str]) -> list[GenericEntity]:
        """
        Retrieve entities by using an accession

        :param accession_list: list of strings containing the accessions.
        :return:
        """
        return self.api.retrieve(accession_list)




from behave import *

import json
import pandas as pd

import sys
sys.path.insert(0, "../../")

from biobroker.output_processor import TsvOutputProcessor, XlsxOutputProcessor, GenericOutputProcessor
from biobroker.metadata_entity import Biosample


@given('an instance of {output_processor} and a {metadata_entity} subclass loaded with the content in {valid_minimal_json}')
def save_method(context, output_processor, metadata_entity, valid_minimal_json):
    context.output_processor_class = eval(output_processor)
    with open(valid_minimal_json, 'r') as f:
        context.metadata_entity = eval(metadata_entity)(json.load(f))


@when('I save the entity as {output_file_path}')
def save_method(context, output_file_path):
    context.output_file_path = output_file_path
    context.output_processor = context.output_processor_class(output_file_path)
    context.output_processor.save([context.metadata_entity])

@then('the output should be equal to {test_file_path}')
def save_method(context, test_file_path):
    kwargs = {}
    match test_file_path:
        case x if x.endswith('tsv'):
            reader_function = pd.read_csv
            kwargs['delimiter'] = '\t'
        case x if x.endswith('xlsx'):
            reader_function = pd.read_excel
            kwargs['engine'] = 'openpyxl'
        case _:
            assert False, "Please create test for this file extension"

    dataframe_from_test = reader_function(context.output_file_path, **kwargs)
    dataframe_from_assets = reader_function(test_file_path, **kwargs)
    assert len(dataframe_from_test.columns.values) == len(dataframe_from_assets.columns.values), "Different number of columns"
    colnames = set(dataframe_from_test.columns.values) | set(dataframe_from_assets.columns.values)
    differences = [dataframe_from_assets[column].values.sort() == dataframe_from_test[column].values.sort() for column in colnames]
    assert all(differences), "Values are different"


import inspect
import logging


class MandatoryFunctionNotSet(Exception):
    """Mandatory function has not been overriden in subclass"""
    def __init__(self, logger: logging.Logger):
        previous_frame = inspect.currentframe().f_back
        caller_function = previous_frame.f_code.co_name
        caller_function_parameters = previous_frame.f_code.co_varnames[1:previous_frame.f_code.co_argcount]
        caller_class = previous_frame.f_locals['self'].__class__.__name__
        message = f"Mandatory function '{caller_function}' not set in '{caller_class}'. Need to set up the following " \
                  f"parameters for the function: {', '.join(caller_function_parameters)}"
        logger.error(message)
        super().__init__(message)


"""
Wrapper submodule. Not a lot to say here - This is just a wrapper to make it easier for users to use basic
library functionality:

- :func:`broker.wrangler.Wrangler.submit`
- Retrieve: :func:`broker.wrangler.Wrangler.retrieve_entities`
- Save: :func:`broker.wrangler.Wrangler.save_results`
"""
from biobroker.wrangler.wrangler import Wrangler

# This lets Sphinx know you want to document package.module.Class as package.Class.
__all__ = ['Wrangler']


from behave import *
import os
import re

import sys
sys.path.insert(0, "../../")

from biobroker.authenticator import WebinAuthenticator, GenericAuthenticator

def load_credentials(prefix):
    username = os.environ.get(f'{prefix}_USERNAME')
    password = os.environ.get(f'{prefix}_PASSWORD')
    assert username and password, f"Username or password missing from environment variables for authenticator with prefix {prefix}"
    return username, password

@given("a {auth} and a set of credentials loaded from '.env' with prefix {prefix}")
def step_impl(context, auth, prefix):
    context.authenticator = eval(auth)
    context.username, context.password = load_credentials(prefix)

@when("I load the instance")
def step_impl(context):
    os.environ['API_ENVIRONMENT'] = "dev"
    context.authenticator_instance = context.authenticator(context.username, context.password)

@then("the auth endpoint should point to development")
def step_impl(context):
    assert "dev" in context.authenticator_instance.base_uri, "Auth endpoint should point to dev for tests"

@then("the token should conform to pattern {pattern}")
def step_impl(context, pattern):
    assert re.match(pattern, context.authenticator_instance.token), "Returned token does not match expected pattern"



from behave import *
import os

import sys
sys.path.insert(0, "../../")

from biobroker.api.exceptions import BiosamplesValidationError
from biobroker.input_processor import XlsxInputProcessor
from biobroker.output_processor import XlsxOutputProcessor
from biobroker.authenticator import WebinAuthenticator
from biobroker.api import BsdApi
from biobroker.metadata_entity import Biosample

import os

def load_credentials(prefix):
    username = os.environ.get(f'{prefix}_USERNAME')
    password = os.environ.get(f'{prefix}_PASSWORD')
    assert username and password, f"Username or password missing from environment variables for authenticator with prefix {prefix}"
    return username, password

@given("an input excel file and all the biobroker entities set up")
def load_everything(context):
    os.environ['API_ENVIRONMENT'] = "dev"
    context.excel_file_path = 'assets/valid_minimal.xlsx'
    context.input_processor = XlsxInputProcessor(context.excel_file_path)
    context.output_processor = XlsxOutputProcessor('output.xlsx')
    context.authenticator = WebinAuthenticator(*load_credentials('WEBIN'))
    context.api = BsdApi(context.authenticator)
    context.samples = context.input_processor.process(Biosample)

@when("I submit using the BsdApi")
def submit(context):
    context.submitted_samples = context.api.submit(context.samples)

@then("the samples should be correctly submitted")
def submit(context):
    assert isinstance(context.submitted_samples, list)
    assert all([isinstance(entity, Biosample) for entity in context.submitted_samples])
    assert all([entity.accession is not None for entity in context.submitted_samples])

@then("I can save it again as an excel file")
def save(context):
    context.output_processor.save(context.submitted_samples)
    assert os.path.isfile('output.xlsx')
    os.remove('output.xlsx')

@when("I add checklist ERC000017 to the sample and submit")
def submit_checklist_error(context):
    context.samples[0]['checklist'] = 'ERC000017'
    try:
        context.api.submit(context.samples)
        assert False, "Shouldn't be allowed to be submitted"
    except Exception as e:
        context.error = e

@then("the sample should raise errors related to the checklist")
def submit_checklist_error(context):
    assert isinstance(context.error, BiosamplesValidationError)

@when("I modify the metadata to contain the mandatory fields for the checklist")
def submit_checklist_pass(context):
    context.samples[0]['project name'] = "Your fake project"
    context.samples[0]['organism'] = "Homo sapiens"
    context.samples[0]['collection date'] = "2024-09-01"
    context.samples[0]['geographic location (country and/or sea)'] = "Mediterranean Sea"
    context.samples[0]['geographic location (latitude)'] = 1.2234
    context.samples[0]['geographic location (longitude)'] = 7.21
    context.samples[0]['broad-scale environmental context'] = "United Kingdom weather"
    context.samples[0]['local environmental context'] = "Mostly rainy"
    context.samples[0]['environmental medium'] = "Please read my plant"


from behave import *

import json

import sys
sys.path.insert(0, "../../")

from biobroker.input_processor import TsvInputProcessor, XlsxInputProcessor, GenericInputProcessor
from biobroker.metadata_entity import Biosample


@given('an instance of {input_processor} loaded with an {input_file}')
def step_load(context, input_processor, input_file):
    context.input_processor = eval(input_processor)(input_file)

@when('I have a JSON file with the same content')
def step_load_json(context):
    with open('assets/valid_minimal.json', 'r') as f:
        context.valid_minimal_json = json.load(f)

@then('the input_data must be the same as the JSON file')
def compare(context):
    assert context.input_processor.input_data[0] == context.valid_minimal_json

@when('I call the transform function on the data with a map')
def transform(context):
    context.input_processor.transform({'other_field': 'new_field'})

@then("the input_data key 'other_field' must be named 'new_field'")
def transform(context):
    assert context.input_processor.input_data[0].get('other_field') is None
    assert context.input_processor.input_data[0].get('new_field') is not None


# Feature: Integration

@when('I call the process method providing with a {metadata_entity_class} class')
def process(context, metadata_entity_class):
    context.entity_class = eval(metadata_entity_class)
    context.processed_data = context.input_processor.process(context.entity_class)[0]

@then('it should result in a valid Metadata Entity')
def process(context):
    assert isinstance(context.processed_data, context.entity_class)
    assert context.processed_data.id is not None
    assert context.processed_data.accession is not None


import os
import requests

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from biobroker.authenticator.exceptions import WrongUserOrPassword, UsernameNotValid
from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.generic.logger import set_up_logger


class GenericAuthenticator:
    """
    Generic authenticator class. Defines the expected functions and properties of all subclasses. Some must be
    overridden.

    :param base_uri: Base URI for the request to retrieve the token.
    :param username: Username required for authentication.
    :param password: Password required for authentication.
    :param verbose: Boolean indicating if the logger should be verbose.
    """
    def __init__(self, base_uri: str, username: str, password: str, verbose: bool = False):
        self.base_uri = base_uri

        self.username = username
        self.password = password
        self.auth_endpoint = ""
        self.logger = set_up_logger(self, verbose=verbose)

    @property
    def auth_endpoint(self):
        """
        Name of the endpoint to authenticate

        :return: string
        """
        return self._auth_endpoint

    @auth_endpoint.setter
    def auth_endpoint(self, endpoint):
        raise MandatoryFunctionNotSet(self.logger)

    @property
    def token(self):
        """
        Token property

        :return: Bearer <token>
        """
        return f"Bearer {self._token}"

    @token.setter
    def token(self, _):
        """
        SUBCLASSES ARE REQUIRED TO OVERRIDE THIS SETTER. A subclass must define this property, with the decorator
        @GenericAuthenticator.token.setter

        :param _:
        :return:
        """
        raise MandatoryFunctionNotSet(self.logger)

    # All functions below are wrappers for requests.get/put/patch/delete/post
    def _request(self, url, method, payload: dict) -> requests.Response:
        """
        Handle all requests. If token is expired, reload token and try again. If this results in another error, it will
        be risen.

        :param url: URL to REQUEST
        :param method: Method for the REQUEST
        :param payload: Optional payload, for POST/PUT/PATCH methods
        :return:
        """
        s = requests.Session()
        retries = Retry(total=5, backoff_factor=1)
        s.mount('https://', HTTPAdapter(max_retries=retries))

        r = s.request(url=url, method=method, json=payload, headers={'Authorization': self.token,
                                                                     'Content-Type': 'application/json'},
                      )

        if r.status_code == 401:
            self.logger.warning(f"{method} request returned status code {r.status_code}. "
                                "Refreshing token and trying again.")
            self.token = (self.username, self.password)
            r = s.request(url=url, method=method, json=payload, headers={'Authorization': self.token,
                                                                         'Content-Type': 'application/json'},
                          )
        return r

    def get(self, url: str) -> requests.Response:
        """
        GET a url

        :param url: URL to GET
        :return: response
        """
        return self._request(url, "GET", {})

    def post(self, url: str, payload: dict) -> requests.Response:
        """
        POST a payload to a URL

        :param url: URL to POST
        :param payload: JSON payload
        :return: response
        """
        return self._request(url, "POST", payload)

    def delete(self, url: str) -> requests.Response:
        """
        DELETE an entity on a URL

        :param url: URL to send the DELETE request
        :return: response
        """
        return self._request(url, "DELETE", {})

    def put(self, url: str, payload: dict) -> requests.Response:
        """
        PUT a payload to an URL

        :param url:
        :param payload:
        :return: response
        """
        return self._request(url, "PUT", payload)

    def patch(self, url: str, payload: dict) -> requests.Response:
        """
        PATCH a payload to a URL

        :param url:
        :param payload:
        :return: response
        """
        return self._request(url, "PATCH", payload)


class WebinAuthenticator(GenericAuthenticator):
    """
    Webin authenticator. Retrieves a token from the Webin service.

    :param username: Webin username. Must start with "Webin-"
    :param password: Webin password.
    """
    def __init__(self, username: str, password: str, verbose: bool = False):
        environment = 'dev' if 'dev' == os.environ.get('API_ENVIRONMENT', '') else ''
        base_uri = "https://www.ebi.ac.uk/ena/submit/webin/auth".replace('www', f"www{environment}")

        super().__init__(base_uri, username, password, verbose)
        self.validate_username(username)
        self.auth_endpoint = "token"
        self.token = ""
        self.logger.info(f"WebinAuthenticator set up successfully. Base uri: {self.base_uri}")

    @GenericAuthenticator.auth_endpoint.setter
    def auth_endpoint(self, endpoint: str):
        """Authentication endpoint setter"""
        self._auth_endpoint = os.path.join(self.base_uri, endpoint)

    @GenericAuthenticator.token.setter
    def token(self, _):
        """
        Token setter for property. Returns a token from the ENA Webin service.

        :param _: Leave blank or pass None
        :return:
        """
        self.logger.info(f"Generating token for user {self.username}")
        r = requests.post(f"{self.auth_endpoint}", json={
            "authRealms": [
                "ENA"
            ],
            "password": self.password,
            "username": self.username
        })
        if r.status_code == 401:
            raise WrongUserOrPassword(self.username, self.password, self.logger)
        else:
            r.raise_for_status()
            self._token = r.text

    # Validation methods
    def validate_username(self, username: str):
        """
        Validate a username. For webin, all usernames start with `Webin-`. Can raise
        :exc:`~biobroker.authenticator.exceptions.UsernameNotValid`

        :param username: Webin username.
        :return:
        """
        if not username.startswith("Webin-"):
            raise UsernameNotValid(username, self.logger)



import json
import re

import pydantic_core
from typing import Any, Type

from pydantic import BaseModel

from biobroker.generic.pydantic_model import BiosampleGeneralModel
from biobroker.metadata_entity.exceptions import (RelationshipInvalidSourceError, RelationshipInvalidTargetError,
                                                  EntityValidationError)
from biobroker.generic.exceptions import MandatoryFunctionNotSet
from biobroker.generic.logger import set_up_logger

# MONKEY PATCHING JSON ENCODER TO MAKE ENTITIES JSON SERIALIZABLE #
def _default(self, obj):
    return getattr(obj.__class__, "to_json", _default.default)(obj)


_default.default = json.JSONEncoder().default
json.JSONEncoder.default = _default


class GenericEntity:
    """
    Generic definition of metadata entity.

    :param metadata_content: dictionary with the content of the entity
    :patam data_model: BaseModel subclass determining the metadata model to validate the `metadata_content`.
    """
    def __init__(self, metadata_content: dict, data_model: type[BaseModel], verbose: bool = False):
        self.logger = set_up_logger(self, verbose=verbose)
        self._entity = None
        self.entity = metadata_content
        self.validate(data_model=data_model)

    @property
    def entity(self) -> dict:
        """
        Entity getter.

        :return: self._entity
        """
        return self._entity

    @entity.setter
    def entity(self, metadata_content: dict):
        """
        Setter for the entity property. Must be overridden by subclasses.

        :param metadata_content: Meta
        :return:
        """
        raise MandatoryFunctionNotSet(self.logger)

    @property
    def id(self) -> str:
        """
        id property. Must be overridden by subclasses.

        :return: string with the ID of the entity
        """
        raise MandatoryFunctionNotSet(self.logger)

    @property
    def accession(self) -> str:
        """
        accession property. Must be overridden by subclasses.

        :return: string with the accession of the entity
        """
        raise MandatoryFunctionNotSet(self.logger)

    def validate(self, data_model: Type[BaseModel]):
        """
        Validate the metadata content using a Pydantic data model. Each subclass can define its own data model, or it
        can be provided by the user on validation.
        """
        try:
            self.entity = json.loads(data_model(**self.entity).model_dump_json(exclude_unset=True, by_alias=True))
        except pydantic_core.ValidationError as pydantic_error:
            raise EntityValidationError(self.logger, entity_id=self.id, errors=pydantic_error.errors()) from None

    def flatten(self):
        """
        Flatten the .entity, returning a non-nested dictionary.
        """
        raise MandatoryFunctionNotSet(self.logger)

    def __setitem__(self, key: str, value: str):
        """
        Must be overriden by subclasses.
        """
        raise MandatoryFunctionNotSet(self.logger)

    def __delitem__(self, key: str):
        """
        Must be overriden by subclasses.
        """
        raise MandatoryFunctionNotSet(self.logger)

    def __getitem__(self, item: str):
        """
        Must be overriden by subclasses.
        """
        raise MandatoryFunctionNotSet(self.logger)

    def __contains__(self, item):
        """
        Must be overriden by subclasses.
        """
        raise MandatoryFunctionNotSet(self.logger)

    @staticmethod
    def guidelines() -> str:
        """
        Return a printable string with guidelines on how to fill out each entity subclass. Subclasses are not required
        to override this property, but... it does help, so try to be nice :)

        :return: empty string. This is a generic class and should never be used as is!
        """
        return ""

    # DECODER FUNCTIONS - DEFAULTS STR
    def to_json(self):
        return json.loads(json.dumps(self.entity, default=str))


class Biosample(GenericEntity):
    """
    Biosamples metadata entity. Contains the necessary information to process a non-nested JSON into a valid Biosamples
    sample.

    Current known issue: Only setting up and expecting one value for the properties inside characteristics. Well, this
    is because... Biosamples also expects that! No clue why properties are defaulting to arrays.

    :param metadata_content: non-nested dictionary containing the metadata for the sample.
    :param data_model: Optional parameter, used to evaluate the metadata content. Defaults to :cls:`~biobroker.generic.pydantic_model.BiosampleGeneralModel`
    :param delimiter: optional parameter, used for key delimiters. Used mainly to manage attributes tags, such as
                      'unit' and 'ontologyTerms'. Explained further in
                      :func:`~broker.metadata_entity.biosample.Biosample.__setitem__`, point 4.
    :param verbose: True if logger should be set to INFO. Default WARNING.
    """
    ROOT_PROPERTIES = ['name', 'release', 'relationships', 'accession', 'sraAccession', 'webinSubmissionAccountId',
                       'status', 'update', 'characteristics', 'submittedVia', 'create', '_links', 'submitted', 'taxId',
                       'structuredData', 'externalReferences', 'organization']
    VALID_TAGS = ['text', 'ontologyTerms', 'unit', 'tag']
    VALID_RELATIONSHIPS = ["derived_from", "same_as"]
    EXTERNAL_REFERENCE_FIELD = "url"
    ORGANIZATION = "organizationName"
    def __init__(self, metadata_content: dict, data_model: Type[BaseModel] = BiosampleGeneralModel,
                 delimiter: str = "||", verbose: bool = False):
        self.delimiter = delimiter
        super().__init__(metadata_content, data_model=data_model, verbose=verbose)

    @property
    def id(self):
        """
        Return the property 'id', extracting it from the 'name' property. Defaults to an empty string.

        :return:
        """
        return self.entity.get('name', '')

    @property
    def accession(self) -> str:
        """
        Return the property 'accesssion', extracting it from the 'accession'. Defaults to an emtpy string

        :return:
        """
        return self.entity.get('accession', '')

    @GenericEntity.entity.setter
    def entity(self, metadata: dict):
        """
        Setter for the 'entity' property. Sets up a new sample, with the basic 'name' and 'characteristics' properties.

        :param metadata: non-nested dictionary containing the metadata for the sample.
        """
        self._entity = {"characteristics": {}}
        for field, value in metadata.items():
            if value is None:
                continue
            self[field] = value

    def flatten(self) -> dict:
        """
        Flatten the :attr:`~Biosample.entity` property and return a non-nested dictionary. This will be mostly used for
        output generation.

        :return: flattened dictionary
        """
        sample_json = self.to_json()
        flattened_json = {}
        for key, value in sample_json.items():
            match key:
                case 'relationships':
                    # Relationships are flattened using just the relationship type. Keep it user-friendly!
                    flattened_json = self._flatten_relationships(flattened_json, value)
                case 'characteristics':
                    flattened_json = self._flatten_characteristics(flattened_json, sample_json[key])
                case '_links':
                    # No need to flatten the _links. _links are not useful in output generation.
                    pass
                case 'externalReferences':
                    flattened_json = self._flatten_urls(flattened_json, sample_json[key])
                case _:
                    flattened_json[key] = value
        return flattened_json

    def __getitem__(self, item) -> str | int | dict:
        """
        Special method to get values from the Biosample.entity. Tries to obtain it from root and then characteristics;
        raises ValueError if not found.

        :param item: Value of the key to look up for

        :return: Value of the item if found.
        """
        return self.entity.get('characteristics', {}).get(item, [{}])[0] or self.entity.get(item)

    def __delitem__(self, key: str):
        """
        Special method to delete the values from the Biosample.entity.
        You can delete tags by using delimiter e.g. temperature||unit

        :param key: Key to search for for deletion
        :return:
        """
        if key in Biosample.ROOT_PROPERTIES:
            del self.entity[key]
        else:
            keys = key.split(self.delimiter)
            match len(keys):
                case 1:
                    del self.entity['characteristics'][keys[0]]
                case 2:
                    del self.entity['characteristics'][keys[0]][0][keys[1]]
                case _:
                    # In Biosamples, nesting is limited to 1 level
                    raise KeyError(key)

    def __setitem__(self, key: str, value: Any):
        """
        This is a special method to set up values to "entity" in a dict-like manner. Each entity can decide to implement
        checks (Depending on the archive needs) or just return "self.entity[key] = value".

        Tags for the attributes can be set up from the flattened input dictionary. As such, each sample has
        a delimiter set up, and if it's detected, instead of replacing the value, it will add the tag (e.g.
        `{'size': 1, 'size||unit': 'cm'}` will be translated to `{'size': [{'text': 1, 'unit': cm}]}`). Please see
        https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_sample for more information.

        :param key: name of the attribute.
        :param value: value of the attribute.
        """

        # Root values
        if key in Biosample.ROOT_PROPERTIES:
            self.entity[key] = value
        # Relationships
        elif self.accession and key in Biosample.VALID_RELATIONSHIPS and all([
            self.check_accession(accession) for accession in value.split(self.delimiter)]):
            for target in value.split(self.delimiter):
                self.add_relationship(source=self.accession,
                                      target=target,
                                      relationship=key)
        # External references
        elif key == Biosample.EXTERNAL_REFERENCE_FIELD:
            for url in value.split(self.delimiter):
                self.add_external_reference(url=url)
        # Organizations - Since there is no documentation I will assume it only has a name
        elif key == Biosample.ORGANIZATION:
            for organization in value.split(self.delimiter):
                self.add_organization(organization=organization)
        # characteristics
        else:
            characteristic = self.entity.get('characteristics').get(key, [{}])[0]
            # Managing the elements of the attributes in characteristics. Check
            # https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_sample
            if self.delimiter in key:
                key, tag_name = key.split(self.delimiter)
                if not self._tag_is_valid(tag_name):
                    self.logger.warning(f"Tag '{tag_name}' on property '{key}' is not a documented valid tag. "
                                        "It may be rejected on submission.")
                characteristic = self.entity['characteristics'].get(key, [{}])[0]
                # Kinda hate to have soooo many ifs, but every field has its own rules...
                characteristic[tag_name] = value if tag_name != 'ontologyTerms' else value.split(self.delimiter)
            else:
                characteristic['text'] = value

            self.entity['characteristics'][key] = [characteristic]

    def __contains__(self, item: str) -> bool:
        """
        Special method to check if Biosample.entity  contains 'item'.
        :param item: value of the key to check for.
        :return: True if found, False if not found.
        """
        return item in self.entity or item in self.entity.get('characteristics')

    def add_relationship(self, source, target, relationship):
        """
        Add a relationship to the entity. Source must be the entity's accession; target must be a valid BioSamples
        accession; relationship must be a valid relationship.

        :param source: source entity accession (Must be equal to entity)
        :param target: target sample accession
        :param relationship: Relationship between source and target. Valid relationships: :attr:`~Biosample.VALID_RELATIONSHIPS`
        """
        if not self.accession == source:
            raise RelationshipInvalidSourceError(logger=self.logger, source=source, sample_id=self.id)
        if not self.check_accession(target):
            raise RelationshipInvalidTargetError(logger=self.logger, target=target, sample_id=self.id)

        if 'relationships' not in self.entity:
            self.entity['relationships'] = []
        self.entity['relationships'].append({
            "source": source,
            "target": target,
            "type": relationship
        })

    def add_external_reference(self, url: str):
        """
        Add an external reference to the entity.

        :param url: URL to the external reference
        """
        if 'externalReferences' not in self.entity:
            self.entity['externalReferences'] = []
        self.entity['externalReferences'].append({'url': url})

    def add_organization(self, organization):
        if 'organization' not in self.entity:
            self.entity['organization'] = []
        self.entity['organization'].append({'Name': organization})

    def _flatten_relationships(self, flattened_json: dict, relationships: list[dict]) -> dict:
        """
        Flatten relationships. To make it user friendly, limit it to Biosample --> target relationships. Biosamples
        defines both directionalities, but for flattening, this is way simpler and no information is lost.

        :param flattened_json: Flattened dictionary in progress.
        :param relationships: list of relationship dictionaries.

        :return: flattened dictionary with the processed relationships incorporated.
        """
        source_accession = self.accession
        relationship_types = set(Biosample.VALID_RELATIONSHIPS)
        for relationship in relationships:
            if relationship['source'] != source_accession:
                continue
            if relationship['type'] not in flattened_json:
                flattened_json[relationship['type']] = []
            flattened_json[relationship['type']].append(relationship['target'])
        for relationship_type in relationship_types:
            flattened_json[relationship_type] = "||".join(flattened_json.get(relationship_type, []))
        return flattened_json

    def _flatten_characteristics(self, flattened_json: dict, characteristics: dict) -> dict:
        """
        flatten the characteristics.
        :param flattened_json: Flattened dictionary in progress.
        :param characteristics: dictionary with the characteristics.

        :return: flattened dictionary with the processed characteristics incorporated.
        """
        for field_name, values in characteristics.items():
            characteristic = values[0]
            for tag, value in characteristic.items():
                match tag:
                    case "text":
                        flattened_json[field_name] = value
                    case _:
                        # Returning units and ontologyTerm's to their natural habitat
                        flattened_json[f"{field_name}{self.delimiter}{tag}"] = value
            flattened_json[field_name] = "||".join([f"{v['text']}" for v in values])
        return flattened_json

    def _flatten_urls(self, flattened_json: dict, urls: dict):
        """
        Flatten the externalReferences.

        :param flattened_json: Flattened dictionary in progress.
        :param urls: dictionary with the urls.

        :return: flattened dictionary with the processed urls incorporated.

        """
        flattened_json['url'] = self.delimiter.join([value['url'] for value in urls])
        return flattened_json

    @staticmethod
    def check_accession(accession) -> bool:
        """
        Check if the provided accession conforms to a BioSamples identifier. Pattern extracted from
        https://registry.identifiers.org/registry/biosample#!

        :param accession: Accession ID for the sample in BioSamples
        :return: True if correct format, False if not
        """
        return True if re.match('^SAM[NED](\\w)?\\d+$', accession) else False

    @staticmethod
    def _tag_is_valid(tag: str) -> bool:
        """
        Check if a tag is valid. Tags are evaluated against the :attr:`~Biosample.VALID_TAGS` global. VALID_TAGS extracted from:
        https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_sample

        :param tag: string with the tag name
        :return: True if valid, False if invalid
        """
        return tag in Biosample.VALID_TAGS

    @staticmethod
    def guidelines() -> str:
        """
        Guidelines for filling out sample metadata for BioSamples.

        :return: Printable string with guidelines.
        """
        return BIOSAMPLES_GUIDELINES


BIOSAMPLES_GUIDELINES = "A Biosamples entity MUST have the following properties set:\n" \
                        "\t- name: a descriptive title for the sample\n" \
                        "\t- organism: a string that validates against NCBITaxon records \n" \
                        "\t- release: date of release for the metadata of the entity, in YYYY-MM-DD format. Accepts iso format\n" \
                        "For more information, please see " \
                        "https://www.ebi.ac.uk/biosamples/docs/references/api/submit#_submission_minimal_fields.\n\n" \
                        "To indicate relationships in the samples, please use a field named after the relationship" \
                        "itself: namely, 'derived_from', 'same_as', 'has_member' or 'child_of'.\nPlease see" \
                        "https://www.ebi.ac.uk/biosamples/docs/guides/relationships"


from behave import *
import os
import json

import sys
sys.path.insert(0, "../../")


from biobroker.authenticator import WebinAuthenticator
from biobroker.api import BsdApi
from biobroker.api.exceptions import StructuredDataError
from biobroker.metadata_entity import Biosample

import mock

def load_accessioned_entity(file_name, times, metadata_entity):
    with open(f"assets/{file_name}") as f:
        content = metadata_entity(json.load(f))
    return content if times == 1 else [content] * times

def load_credentials(prefix):
    username = os.environ.get(f'{prefix}_USERNAME')
    password = os.environ.get(f'{prefix}_PASSWORD')
    assert username and password, f"Username or password missing from environment variables for authenticator with prefix {prefix}"
    return username, password

@given("all the API classes and their corresponding authenticator")
def preload_instances(context):
    context.instances = {}
    for subclass_api, subclass_authenticator, prefix in (('BsdApi', 'WebinAuthenticator', 'WEBIN'),):
        subclass_api_object = eval(subclass_api)
        subclass_authenticator_object = eval(subclass_authenticator)
        os.environ['API_ENVIRONMENT'] = "dev"
        subclass_authenticator_instance = subclass_authenticator_object(*load_credentials(prefix))
        context.instances[subclass_api] = subclass_api_object(subclass_authenticator_instance)
        assert "dev" in context.instances[subclass_api].base_uri


@given("{length} {metadata_entity} filled with content from {metadata_json_path}")
def submit(context, length, metadata_entity, metadata_json_path):
    context.length = int(length)
    context.metadata_entity_subclass = eval(metadata_entity)
    with open(metadata_json_path, 'r') as f:
        context.entity_metadata_json = json.load(f)
    context.metadata_entity = [eval(metadata_entity)(context.entity_metadata_json)] * context.length

@when("an API instance named {api_instance_name} is used to submit it to the archive")
def submit(context, api_instance_name):
    context.api_subclass_instance = context.instances[api_instance_name]
    function_to_mock = f"biobroker.api.{api_instance_name}._submit"
    if context.length > 1:
        function_to_mock += "_multiple"
    with mock.patch(function_to_mock) as mocked_submit_function:
        accessioned_entities = load_accessioned_entity(f"accessioned_{api_instance_name}_entity.json", context.length,
                                                       context.metadata_entity_subclass)
        # Accessioned entities = [Biosample(accesioned.json) * length]
        mocked_submit_function.return_value = accessioned_entities # [Biosample(accessioned.json)]
        context.submitted_entities = context.api_subclass_instance.submit(context.metadata_entity)

@then("we get a list of entities with the expected length and the accession set up")
def submit(context):
    assert len(context.submitted_entities) == context.length, "Expected length and number of submitted entities does not match"
    for entity in context.submitted_entities:
        assert entity.accession == "SAMEA131439753"

@then("retrieving those entities by accession should result in the exact same entities")
def retrieve(context):
    accessions = [entity.accession for entity in context.submitted_entities]
    retrieved_entities = context.api_subclass_instance.retrieve(accessions)
    assert all([context.submitted_entities[i].entity == retrieved_entities[i].entity
                for i in range(len(retrieved_entities))])

@when("An API instance named {api_instance_name} is used to update the entity")
def update(context, api_instance_name):
    context.api_instance_name = api_instance_name
    context.api_subclass_instance = context.instances[api_instance_name]
    function_to_mock = f"biobroker.api.{api_instance_name}._update"
    if context.length > 1:
        function_to_mock += "_multiple"
    with mock.patch(function_to_mock) as mocked_update_function:
        context.accessioned_entities = load_accessioned_entity(f"accessioned_{api_instance_name}_entity.json", context.length,
                                                       context.metadata_entity_subclass)
        mocked_update_function.return_value = context.accessioned_entities
        context.updated_entities = context.api_subclass_instance.update(context.metadata_entity)

@then("we get a list of entities with the expected length and the updated metadata")
def update(context):
    assert len(context.updated_entities) == context.length
    original_acessioned = context.accessioned_entities if isinstance(context.accessioned_entities, list) else [context.accessioned_entities]
    assert all([original_acessioned[i] is context.updated_entities[i] for i in range(len(context.updated_entities))])

@given("an invalid structured data object")
def struc_data_negative(context):
    with open('assets/structured_data_invalid.json', 'r') as f:
        context.structured_data = json.load(f)

@when("the structured data is submitted to Biosamples")
def struc_data_negative(context):
    biosamples_api = context.instances['BsdApi']
    try:
        biosamples_api.submit_structured_data(context.structured_data)
        assert False, "This method call should raise an error"
    except StructuredDataError as e:
        context.error = e

@then("it should raise an error with the expected error messages")
def struc_data_negative(context):
    assert "accession" in context.error.message
    assert "data-->0-->webinSubmissionAccountId" in context.error.message
    assert "data-->0-->content-->0-->resistancePhenotype" in context.error.message
    assert "data-->0-->content-->0-->platform" in context.error.message

@given("a valid structured data object with an invalid accession")
def struc_data_invalid_acc(context):
    with open('assets/structured_data_invalid_accession.json', 'r') as f:
        context.structured_data = json.load(f)

@then("it should raise an error regarding the accession")
def struc_data_invalid_acc(context):
    assert "accession: Value provided must match pattern" in context.error.message
    assert context.error.message.count('\n\t-') == 1, "Only accession error should be raised"


"""
Metadata entities. The goal of this module is to define the metadata entities that each archive uses. All the metadata
is saved under the self.entity property, but the whole class can be accessed as a dictionary.

To make it easier for the users, it is expected that all properties can be accessed as a root property.
The subclasses then must define how properties are accessed by defining "__setitem__", "__getitem__" and "__contains__".
If unsure, look at the "Biosample" subclass to understand how this is implemented.

The base generic entity also defines a custom JSON encoder, to avoid JSON.dumps issues. Any class not serializable will
be turned to string.

**Mandatory arguments**:

    - metadata_content: Metadata content for the entity, in JSON format.
    - field_mapping: Map <archive_key_name>:<metadata_content_key_name> in JSON format
    - keep_non_mapped_fields: Boolean, wether to keep the non-mapped fields.

**Optional arguments**:

- verbose: set to `True` if you want `INFO` and above-level logging events. If not set or set to False, only `WARNING`
  and above will be displayed

**Subclasses of GenericEntity must define the following methods/properties**:

- @GenericEntity.setter
- id property
- accession property
- validate
- flatten
- dictionary special methods ('__setitem__', '__delitem__', '__getitem__', '__contains__')

**Subclasses of GenericEntity SHOULD define the following methods/properties**:

- guidelines

**Aspects to improve**:

- Biosamples entity: taxonId or organism must be set up. Currently allows entities to be created without those fields.
- Biosamples entity: Currently the root keys are hardcoded in the submodule. I wonder what would be the best way to
  indicate them without complicating the code and depending too much on external files. These are not bound to change
  much - but still, not good practice to have them in the code.
"""

from .metadata_entity import GenericEntity, Biosample

__all__ = ['GenericEntity', 'Biosample']



import logging


class UsernameNotValid(Exception):
    """Username is not deemed valid for the authenticator"""
    def __init__(self, username: str, logger: logging.Logger):
        message = f"Bad username provided: '{username}'. Maybe you meant to use another authenticator?"
        logger.error(message)
        super().__init__(message)


class WrongUserOrPassword(Exception):
    """Wrong username or password provided to authenticator"""
    def __init__(self, username: str, password: str, logger: logging.Logger):
        message = f"Bad user or password: username={username}, password={password}"
        logger.error(message)
        super().__init__(message)


class DomainAlreadyExists(Exception):
    """Domain already exists"""
    def __init__(self, domain: str, logger: logging.Logger):
        message = f"Domain '{domain}' already exists in AAP. Please try another name"
        logger.error(message)
        super().__init__(message)



"""
Shared exceptions, logging and pydantic models. Not a lot to report here.

Pydantic data models may contain more classes as validator for fields; for stand-alone models,
use the classes marked with `Model` at the end of the class name.
"""

